import multiprocessing
from dataclasses import dataclass
from enum import Enum
from typing import Protocol, Callable, List, Optional, Any

from torch import nn, dtype
from torch.optim import Optimizer
from torch.utils.data.dataloader import default_collate

from xztrainer.logger import LoggingEngineConfig
from xztrainer.logger.stream import StreamLoggingEngineConfig


class SchedulerType(Enum):
    STEP = 'step'
    EPOCH = 'epoch'


class CheckpointType(Enum):
    MODEL_ONLY = 'model_only'
    XZTRAINER = 'xztrainer'


class LRSchedulerProtocol(Protocol):
    def step(self):
        ...

    def state_dict(self):
        ...

    def load_state_dict(self, state_dict):
        ...


@dataclass
class XZTrainerConfig:
    batch_size: int
    batch_size_eval: int
    epochs: int
    optimizer: Callable[[nn.Module], Optimizer]
    amp_dtype: Optional[dtype] = None
    experiment_name: str = 'master'
    gradient_clipping: float = 1.0
    scheduler: Optional[Callable[[Optimizer, int], LRSchedulerProtocol]] = None
    scheduler_type: Optional[SchedulerType] = None
    dataloader_num_workers: int = multiprocessing.cpu_count()
    dataloader_pin_memory: bool = True
    dataloader_persistent_workers: bool = True
    dataloader_shuffle_train_dataset: bool = True
    accumulation_batches: int = 1
    print_steps: int = 100
    eval_steps: int = 0
    save_steps: int = 100
    save_keep_n: int = -1
    save_dir: str = 'checkpoint'
    collate_fn: Callable[[List[object]], Any] = default_collate
    logger: LoggingEngineConfig = StreamLoggingEngineConfig()



from types import TracebackType
from typing import Optional, Type, List

from . import LoggingEngine, LoggingEngineConfig, ClassifierType


class ComposeLoggingEngine(LoggingEngine):
    def __init__(self, experiment_name: str, config: 'ComposeLoggingEngineConfig'):
        super().__init__()

        self.engines = [engine.create_engine(experiment_name) for engine in config.engines]

    def log_scalar(self, classifier: ClassifierType, value: float):
        for engine in self.engines:
            engine.log_scalar(classifier, value)

    def flush(self):
        for engine in self.engines:
            engine.flush()

    def update_time_step(self, time_step: int):
        super().update_time_step(time_step)
        for engine in self.engines:
            engine.update_time_step(time_step)

    def update_top_classifier(self, classifier: ClassifierType):
        super().update_top_classifier(classifier)
        for engine in self.engines:
            engine.update_top_classifier(classifier)

    def __enter__(self) -> 'LoggingEngine':
        return self

    def __exit__(self, __exc_type: Optional[Type[BaseException]], __exc_value: Optional[BaseException], __traceback: Optional[TracebackType]):
        for engine in self.engines:
            engine.__exit__(__exc_type, __exc_value, __traceback)


class ComposeLoggingEngineConfig(LoggingEngineConfig):
    def __init__(self, *engines: LoggingEngineConfig):
        self.engines = engines

    def create_engine(self, experiment_name: str) -> ComposeLoggingEngine:
        return ComposeLoggingEngine(experiment_name, self)



import random
from typing import Sized, Iterator, List, Optional, Any, Dict

from torch.utils.data import Sampler


class ReusableSequentialSampler(Sampler[int]):
    _indices: List[int]

    def __init__(self, indices: List[int], shift_steps: int) -> None:
        super().__init__(indices)
        self._indices = indices
        self._shift_steps = shift_steps

    def __iter__(self) -> Iterator[int]:
        return iter(self._indices[self._shift_steps:])

    def __len__(self) -> int:
        return max(len(self._indices) - self._shift_steps, 0)

    def save_state(self, batch_i: int, batch_size: int) -> Dict[str, Any]:
        return {
            'indices': self._indices,
            'shift_steps': (batch_i + 1) * batch_size
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]):
        return cls(state['indices'], state['shift_steps'])

    @classmethod
    def new(cls, data: Sized, shuffle: bool):
        indices = list(range(len(data)))
        if shuffle:
            random.shuffle(indices)
        return cls(indices, 0)



from abc import ABCMeta, abstractmethod
from contextlib import AbstractContextManager
from types import TracebackType
from typing import Union, Iterable, Optional, Type

ClassifierType = Union[str, Iterable[str]]


def convert_classifier(classifier: ClassifierType) -> Iterable[str]:
    if isinstance(classifier, str):
        return classifier,
    elif isinstance(classifier, Iterable):
        return tuple(classifier)
    else:
        raise ValueError(f'Invalid classifier type: {type(classifier)}')


class LoggingEngine(AbstractContextManager, metaclass=ABCMeta):
    def __init__(self):
        self._time_step = 0
        self._top_classifier = ()

    @abstractmethod
    def log_scalar(self, classifier: ClassifierType, value: float):
        ...

    @abstractmethod
    def flush(self):
        ...

    def update_time_step(self, time_step: int):
        self._time_step = time_step

    def update_top_classifier(self, classifier: ClassifierType):
        self._top_classifier = convert_classifier(classifier)

    def __enter__(self) -> 'LoggingEngine':
        return self

    def __exit__(self, __exc_type: Optional[Type[BaseException]], __exc_value: Optional[BaseException],
                 __traceback: Optional[TracebackType]):
        pass


class LoggingEngineConfig(metaclass=ABCMeta):
    @abstractmethod
    def create_engine(self, experiment_name: str) -> LoggingEngine:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['xztrainer', 'xztrainer.logger']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.22.1', 'tqdm>=4.62.3']

extras_require = \
{'deepspeed:sys_platform == "linux"': ['deepspeed>=0.5.10'],
 'tensorboard': ['tensorboard>=2.8.0'],
 'torch': ['torch>=1.10.0']}

setup_kwargs = {
    'name': 'xztrainer',
    'version': '0.5.1',
    'description': 'A customizable training pipeline for PyTorch',
    'long_description': 'None',
    'author': 'Maxim Afanasyev',
    'author_email': 'mr.applexz@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
